export function point(basis, x, y) {
  basis._context.bezierCurveTo(
    (2 * basis._x0 + basis._x1) / 3,
    (2 * basis._y0 + basis._y1) / 3,
    (basis._x0 + 2 * basis._x1) / 3,
    (basis._y0 + 2 * basis._y1) / 3,
    (basis._x0 + 4 * basis._x1 + x) / 6,
    (basis._y0 + 4 * basis._y1 + y) / 6
  );
};

function basis(context) {
  return new Basis(context);
}

function Basis(context) {
  this._context = context;
}

Basis.prototype = {
  lineStart: function() {
    this._x0 = this._x1 =
    this._y0 = this._y1 = NaN;
    this._state = 0;
  },
  lineEnd: function() {
    switch (this._state) {
      case 1: this._context.closePath(); break;
      case 3: point(this, this._x1, this._y1); // proceed
      case 2: this._context.lineTo(this._x1, this._y1); break;
    }
  },
  point: function(x, y) {
    x = +x, y = +y;
    switch (this._state) {
      case 0: this._state = 1; this._context.moveTo(x, y); break;
      case 1: this._state = 2; this._context.lineTo((5 * this._x1 + x) / 6, (5 * this._y1 + y) / 6); break;
      case 2: this._state = 3; // proceed
      default: point(this, x, y); break;
    }
    this._x0 = this._x1, this._x1 = x;
    this._y0 = this._y1, this._y1 = y;
  }
};

export default basis;
